import osmnx as ox
import networkx as nx
import folium
import itertools

from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.ops import linemerge


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)

# read graph
#G = ox.io.load_graphml(filepath='e:/biz/graph/gila_trails.gml')
#ox.distance.add_edge_lengths(G)

#location where you want to find your route
place     = 'Gila National Forest, New Mexico, United States'
# Using network_type='walk' seems to include dirt roads, so use a custom filter
cf = '["highway"~"path|footway"]'
# create graph from OSM within the boundaries of some 
# geocodable place(s)
G = ox.graph_from_place(place,simplify=False,retain_all=True,custom_filter=cf)
#G = ox.simplification.simplify_graph(G, strict=False, remove_rings=True, track_merged=True)


G = G.to_undirected() # we prefer to work with an undirected multigraph



def draw(Q,title='',show_nodes = False):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs
    
    nodes, edges = ox.graph_to_gdfs(QD)
    m = edges.explore(color='#ff00ff', tiles="openstreetmap")
    if show_nodes:    
        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    


    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
    folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    m.save("e:/biz/graph/map_" + title + ".html")
    m.show_in_browser()
    print('*************', flush=True)
    print("", flush=True)
 
  
def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('')
        
report(G)
draw(G, show_nodes=True)

        

# get undirected multigraph
H = ox.utils_graph.get_undirected(G)
report(H,'in undirected multigraph')

# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(H), key=len)
J = nx.subgraph(H, nodes).copy()
report(J,'in largest subgraph')
draw(J, 'largest subgraph')

# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

delete_ends(J)
report(J,'in undirected multigraph with ends removed')
draw(J,'Ends Removed')

# from https://stackoverflow.com/questions/65568844/typeerror-unhashable-type-linestring-when-using-ox-simplify-graph
#def simplify_graph_modified(G, strict=True, remove_rings=True):



T = J.to_directed()
T = ox.simplification.simplify_graph(T, strict=True, remove_rings=True, track_merged=True)
J = T.to_undirected()
            
report(J,'with two-nodes merged')
draw(J, 'Two-nodes merged', show_nodes=True)
        
# get a graph of just the odd nodes, improve later to handle 4-way intersections

def odd_nodes_graph(Q):
    R = Q.copy()
    unwanted_edges = []
    for u, v, d in R.edges(data=True):
        if R.degree(u) % 2 != 1 or R.degree(v) % 2 != 1:
            unwanted_edges.append( (u,v) )
    R.remove_edges_from( unwanted_edges )
    return R
    

K = odd_nodes_graph(J)           
report(K,'in odd-nodes graph')
draw(K,'Odd-nodes Graph')



#print(all_maximal_matchings(K))
#all_matchings = all_maximal_matchings(K)

# add in edges that bypass 4-nodes in graph
# fix: 4-way connected to a 4-way
# fix: 4-way connected to a 4-way
def complete_3node_graph(Q,R):
    n4list = []
    for node, neighbors in Q.adjacency():
        if len(neighbors) == 4:
            n4list.append( (node,neighbors) )
            nlist = []
            llen = []
            for neighbor, edict in neighbors.items():
                w = min([d['length'] for d in edict.values()])
                nlist.append(neighbor)
                llen.append(w)
              #  for dic in edict.items():
                    #w = dic['length']
                    #nlist.append(neighbor)
                    #llen.append(w)
            # could use itertools?
            for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:
                R.add_edge(nlist[i],nlist[j], length= llen[i]+llen[j]) 
    return n4list

k_fix_4_list = complete_3node_graph(J,K)

report(K,'in odd-nodes 4node bypass')
# draw will not work on odd-nodes 4node bypass, since some edges do not have geometry attr
### #draw(K,'odd-nodes 4node bypass', 
 
# get a minimum matching of odd nodes
odd_matches_set = nx.algorithms.min_weight_matching(nx.Graph(K),'length')
# perhaps a matching is not possible?
# Might have to do my own algorithm?
print('number of odd matches', len(odd_matches_set))

# delete matched edges
# do I need to de-duplicate odd_matches_list?
K.remove_edges_from(odd_matches_set)

report(K,'with odd-node matches removed')
draw(K,'Odd-node Matches Removed')

# https://stackoverflow.com/questions/51933046/python-networkx-get-unique-matching-combinations#51934312
# This function is a thing of beauty, way beyond my abilities.
# I tried building my own, with much wailing, before finding this concise gem.
def all_maximal_matchings(T):

    maximal_matchings = []
    partial_matchings = [{(u,v)} for (u,v) in T.edges()]

    while partial_matchings:
        # get current partial matching
        m = partial_matchings.pop()
        nodes_m = set(itertools.chain(*m))

        extended = False
        for (u,v) in T.edges():
            if u not in nodes_m and v not in nodes_m:
                extended = True
                # copy m, extend it and add it to the list of partial matchings
                m_extended = set(m)
                m_extended.add((u,v))
                partial_matchings.append(m_extended)

        if not extended and m not in maximal_matchings:
            maximal_matchings.append(m)
            print('matching ', len(maximal_matchings), ": ", m)

    return maximal_matchings


# how do I fix the 4-nodes?
def restore_4node_graph(Q,fix_4_list):
    for node, neighbors in fix_4_list:
       i = 0
       nlist = []
       vlist = []
       for neighbor, edict in neighbors.items():
           # for dic in edict:
           nlist.append(neighbor)
           vlist.append(edict[0])
       for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:               
            if Q.has_edge(nlist[i],nlist[j]):
                Q.remove_edge(nlist[i],nlist[j])
                if not Q.has_edge(node,nlist[i]):
                    Q.add_edge(node,nlist[i],**vlist[i]) #fix
                if not Q.has_edge(node,nlist[j]):                                  
                    Q.add_edge(node,nlist[j],**vlist[j])

restore_4node_graph(K,k_fix_4_list)

report(K,'odd-matches deleted with 4nodes restored')
#draw(K,'4nodes Restored')


# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(K), key=len)
L = nx.subgraph(K, nodes).copy()
report(L,'in largest subgraph with odd-node matches removed')
draw(L,'Largest Subgraph with Odd-Node Matches Removed')

if nx.is_eulerian(L):
    print('L is eulerian')
    print('The node path is: ', [u for u,v in nx.eulerian_circuit(L)])
            

# convert back to multidigraph
M = L.to_directed()
#M = nx.MultiDiGraph()
#M = nx.compose(M,L)
#MR = L.reverse(copy=True)
#M = nx.compose(M,MR)

#report(M, 'in multigraph')
       

# save
  



